@@ -5,6 +5,7 @@ from pathlib import Path
 from functools import partial, wraps
 
 import torchaudio
+from torchaudio import transforms as T
 from torchaudio.functional import resample
 
 import sounddevice as sd
@@ -17,10 +18,14 @@ from torch.nn import LSTM, GRU, ConvTranspose1d, Module, ModuleList
 
 from numpy import ndarray
 
+import einx
 from einx import add, multiply, divide
 from einops import rearrange, repeat, reduce, pack, unpack
 from einops.layers.torch import Rearrange
 
+import librosa
+import matplotlib.pyplot as plt
+
 # ein tensor notation:
 
 # b - batch
@@ -49,6 +54,16 @@ def divisible_by(num, den):
 def identity(t):
     return t
 
+def is_empty(t: Tensor):
+    return t.numel() == 0
+
+def round_down_to_multiple(num, mult):
+    return (num // mult) * mult
+
+def lens_to_mask(lens: Tensor, max_len):
+    seq = torch.arange(max_len, device = lens.device)
+    return einx.greater('b, n -> b n', lens, seq)
+
 # residual
 
 def residual(fn):
@@ -138,7 +153,7 @@ class STFT(Module):
             return_complex = True
         )
 
-        return stft
+        return stft, stft.abs()
 
 # Tasnet applies a hann window to the convtranspose1d
 
@@ -175,7 +190,7 @@ class HSTasNet(Module):
         *,
         dim = 500,            # they have 500 hidden units for the network, with 1000 at fusion (concat from both representation branches)
         small = False,        # params cut in half by 1 layer lstm vs 2, fusion uses summed representation
-        stereo = False,
+        stereo = True,
         num_basis = 1500,
         segment_len = 1024,
         overlap_len = 512,
@@ -183,7 +198,8 @@ class HSTasNet(Module):
         sample_rate = 44_100, # they use 41k sample rate
         num_sources = 4,      # drums, bass, vocals, other
         torch_compile = False,
-        use_gru = False
+        use_gru = False,
+        norm_before_mask_estimate = True # for some reason, training is unstable without a norm - improvise by adding an RMSNorm before final projection
     ):
         super().__init__()
 
@@ -215,6 +231,10 @@ class HSTasNet(Module):
 
         # spec branch encoder stft hparams
 
+        self.n_fft = n_fft
+        self.win_length = segment_len
+        self.hop_length = overlap_len
+
         self.stft = STFT(
             n_fft = n_fft,
             win_length = segment_len,
@@ -229,6 +249,7 @@ class HSTasNet(Module):
         )
 
         self.to_spec_masks = nn.Sequential(
+            nn.RMSNorm(dim) if norm_before_mask_estimate else nn.Identity(),
             nn.Linear(dim, spec_dim_input * num_sources),
             Rearrange('b n (s f t) -> (b s) f n t', s = audio_channels, t = num_sources)
         )
@@ -245,6 +266,7 @@ class HSTasNet(Module):
         )
 
         self.to_waveform_masks = nn.Sequential(
+            nn.RMSNorm(dim) if norm_before_mask_estimate else nn.Identity(),
             nn.Linear(dim, num_sources * num_basis),
             Rearrange('... (t basis) -> ... basis t', t = num_sources)
         )
@@ -253,11 +275,11 @@ class HSTasNet(Module):
 
         # init mask to identity
 
-        nn.init.zeros_(self.to_spec_masks[0].weight)
-        nn.init.constant_(self.to_spec_masks[0].bias, 1.)
+        nn.init.zeros_(self.to_spec_masks[1].weight)
+        nn.init.constant_(self.to_spec_masks[1].bias, 1.)
 
-        nn.init.zeros_(self.to_waveform_masks[0].weight)
-        nn.init.constant_(self.to_waveform_masks[0].bias, 1.)
+        nn.init.zeros_(self.to_waveform_masks[1].weight)
+        nn.init.constant_(self.to_waveform_masks[1].bias, 1.)
 
         # they do a single layer of lstm in their "small" variant
 
@@ -291,22 +313,75 @@ class HSTasNet(Module):
     def num_parameters(self):
         return sum([p.numel() for p in self.parameters()])
 
+    # get a spectrogram figure based on hparams of the model
+
+    def save_spectrogram_figure(
+        self,
+        save_path: str | Path,
+        audio: ndarray | Tensor,
+        figsize = (10, 4),
+        overwrite = False
+    ):
+
+        if isinstance(save_path, str):
+            save_path = Path(save_path)
+
+        assert overwrite or not save_path.exists(), f'{str(save_path)} already exists'
+
+        # cast to torch tensor
+
+        if isinstance(audio, ndarray):
+            audio = torch.from_numpy(audio)
+
+        assert audio.ndim == 1
+        audio = audio.detach().cpu()
+
+        # stft to magnitude to db
+
+        stft = torch.stft(audio, n_fft = self.n_fft, hop_length = self.hop_length, return_complex = True)
+        magnitude = stft.abs()
+
+        db = T.AmplitudeToDB()(magnitude)
+
+        # pyplot and librosa
+
+        plt.figure(figsize = figsize)
+
+        librosa.display.specshow(
+            db.numpy(),
+            hop_length = self.hop_length,
+            sr = self.sample_rate,
+            x_axis = 's',
+            y_axis = 'hz'
+        )
+
+        plt.colorbar(format = '%+2.0f dB')
+        plt.title('Power/Frequency (dB)')
+        plt.tight_layout()
+
+        plt.savefig(str(save_path))
+
     # saving and loading
 
     def save_tensor_to_file(
         self,
         output_file: str | Path,
         audio_tensor: Tensor,
-        overwrite = False
+        overwrite = False,
+        verbose = False
     ):
         if isinstance(output_file, str):
             output_file = Path(output_file)
 
         assert not exists(output_file) or overwrite, f'{str(output_file)} already exists, set `overwrite = True` if you wish to overwrite'
 
+        if audio_tensor.ndim == 1:
+            audio_tensor = repeat(audio_tensor, 'n -> s n', s = 2 if self.stereo else 1)
+
         torchaudio.save(str(output_file), audio_tensor.cpu(), sample_rate = self.sample_rate)
 
-        print(f'audio saved to {str(output_file)}')
+        if verbose:
+            print(f'audio saved to {str(output_file)}')
 
     def save(self, path, overwrite = True):
         path = Path(path)
@@ -347,6 +422,7 @@ class HSTasNet(Module):
         self,
         device = None,
         return_reduced_sources: list[int] | None = None,
+        auto_convert_to_stereo = True
     ):
         chunk_len = self.overlap_len
         self.eval()
@@ -374,6 +450,11 @@ class HSTasNet(Module):
             if exists(device):
                 audio_chunk = audio_chunk.to(device)
 
+            # auto repeat mono to stereo if model is trained with stereo but received audio is mono
+
+            if auto_convert_to_stereo and self.stereo and audio_chunk.shape[0] == 1:
+                audio_chunk = repeat(audio_chunk, '1 d -> s d', s = 2)
+
             # add past audio chunk
 
             full_chunk = cat((past_audio, audio_chunk), dim = -1)
@@ -406,13 +487,15 @@ class HSTasNet(Module):
         duration_seconds = 10,
         channels = None,
         device = None,
+        auto_convert_to_stereo = True,
         **stream_kwargs
     ):
         assert len(return_reduced_sources) > 0
 
         transform_fn = self.init_stateful_transform_fn(
             return_reduced_sources = return_reduced_sources,
-            device = device
+            device = device,
+            auto_convert_to_stereo = auto_convert_to_stereo
         )
 
         # sounddevice stream callback, where raw audio can be transformed
@@ -447,6 +530,7 @@ class HSTasNet(Module):
         input_file: str | Path,
         return_reduced_sources: list[int],
         output_file: str | Path | None = None,
+        auto_convert_to_stereo = True,
         overwrite = False
     ):
         if isinstance(input_file, str):
@@ -465,17 +549,37 @@ class HSTasNet(Module):
         # curtail to divisible segment lens
 
         audio_len = audio_tensor.shape[-1]
-        rounded_down_len = audio_len // self.segment_len * self.segment_len
+        rounded_down_len = round_down_to_multiple(audio_len, self.segment_len)
 
         audio_tensor = audio_tensor[..., :rounded_down_len]
 
+        # add batch
+
+        audio_tensor = rearrange(audio_tensor, '... -> 1 ...')
+
+        # maybe mono to stereo
+
+        mono_to_stereo = self.stereo and auto_convert_to_stereo and audio_tensor.shape[0] == 1
+
+        if mono_to_stereo:
+            audio_tensor = repeat(audio_tensor, '1 1 n -> 1 s n', s = 2)
+
         # transform
 
         audio_tensor = audio_tensor.to(self.device)
 
         with torch.no_grad():
             self.eval()
-            transformed = self.forward(audio_tensor, return_reduced_sources = return_reduced_sources)
+            transformed, _ = self.forward(audio_tensor, return_reduced_sources = return_reduced_sources)
+
+        # remove batch
+
+        transformed = rearrange(transformed, '1 ... -> ...')
+
+        # maybe stereo to mono
+
+        if mono_to_stereo:
+            transformed = reduce(transformed, 's n -> 1 n', 'mean')
 
         # save output file
 
@@ -484,21 +588,42 @@ class HSTasNet(Module):
 
         assert output_file != input_file
 
-        self.save_tensor_to_file(str(output_file), audio_tensor.cpu(), overwrite = overwrite)
+        self.save_tensor_to_file(str(output_file), transformed.cpu(), overwrite = overwrite)
 
     def forward(
         self,
-        audio,
+        audio,             # (b {s} n)  - {} meaning optional dimension for stereo or not
         hiddens = None,
-        targets = None,
+        targets = None,    # (b t {s} n)
+        audio_lens = None, # (b)
         return_reduced_sources: list[int] | None = None,
-        auto_causal_pad = None
+        auto_causal_pad = None,
+        auto_curtail_length_to_multiple = True,
     ):
         auto_causal_pad = default(auto_causal_pad, self.training)
 
+        assert auto_curtail_length_to_multiple or divisible_by(audio.shape[-1], self.segment_len)
+
+        # take care of audio being passed in that isn't multiple of segment length
+
+        if auto_curtail_length_to_multiple:
+            round_down_audio_len = round_down_to_multiple(audio.shape[-1], self.segment_len)
+            audio = audio[..., :round_down_audio_len]
+
+            if exists(targets):
+                targets = targets[..., :round_down_audio_len]
+
+        # variables
+
         batch, audio_len, device = audio.shape[0], audio.shape[-1], audio.device
 
-        assert divisible_by(audio_len, self.segment_len)
+        # validate lengths
+
+        assert not exists(targets) or audio.shape[-1] == targets.shape[-1]
+
+        assert not is_empty(audio), f'audio is empty, probably insufficient lengthed audio given segment length and auto-rounding down the length'
+
+        # their small version probably does not need a skip connection
 
         maybe_residual = residual if not self.small else identity
 
@@ -514,6 +639,16 @@ class HSTasNet(Module):
 
         assert not (self.stereo and audio.shape[1] != 2), 'audio channels must be 2 if training stereo'
 
+        # handle masking
+
+        need_audio_mask = exists(audio_lens)
+
+        if need_audio_mask:
+            audio_lens = round_down_to_multiple(audio_lens, self.segment_len)
+            assert (audio_lens > 0).all()
+
+            audio_mask = lens_to_mask(audio_lens, audio_len)
+
         # pad the audio manually on the left side for causal, and set stft center False
 
         if auto_causal_pad:
@@ -523,9 +658,7 @@ class HSTasNet(Module):
 
         spec_audio_input = rearrange(audio, 'b s ... -> (b s) ...')
 
-        complex_spec = self.stft(spec_audio_input)
-
-        magnitude = complex_spec.abs()
+        complex_spec, magnitude = self.stft(spec_audio_input)
 
         spec = self.spec_encode(magnitude)
 
@@ -632,7 +765,12 @@ class HSTasNet(Module):
             recon_audio = recon_audio[..., self.causal_pad:]
 
         if exists(targets):
-            recon_loss = F.l1_loss(recon_audio, targets) # they claim a simple l1 loss is better than all the complicated stuff of past
+            recon_loss = F.l1_loss(recon_audio, targets, reduction = 'none' if need_audio_mask else 'mean') # they claim a simple l1 loss is better than all the complicated stuff of past
+
+            if need_audio_mask:
+                recon_loss = rearrange(recon_loss, 'b ... n -> b n ...')
+                recon_loss = recon_loss[audio_mask].mean()
+
             return recon_loss
 
         # outputs
@@ -645,6 +783,9 @@ class HSTasNet(Module):
             next_post_waveform_hidden
         )
 
+        if need_audio_mask:
+            recon_audio = einx.where('b n, b t n,', audio_mask, recon_audio, 0.)
+
         if exists(return_reduced_sources):
             return_reduced_sources = tensor(return_reduced_sources, device = device)
 
