@@ -1,6 +1,7 @@
 from __future__ import annotations
+from shutil import rmtree
 from functools import partial
-from random import random
+from random import random, randrange
 from pathlib import Path
 
 import torch
@@ -8,7 +9,7 @@ from torch import cat, stack, tensor
 from torch.nn import Module
 from torch.optim import Adam
 from torch.optim.lr_scheduler import StepLR
-from torch.utils.data import Dataset, DataLoader
+from torch.utils.data import Dataset, DataLoader, random_split
 from torch.nn.utils.rnn import pad_sequence
 
 import numpy as np
@@ -265,6 +266,7 @@ class Trainer(Module):
         model: HSTasNet,
         dataset: Dataset | MusDB,
         eval_dataset: Dataset | None = None,
+        random_split_dataset_for_eval_frac = 0., # if set higher than 0., will split off this fraction of the training dataset for evaluation - eval_dataset must be None
         optim_klass = Adam,
         batch_size = 128,
         grad_accum_every = 1,
@@ -279,6 +281,7 @@ class Trainer(Module):
         ema_kwargs: dict = dict(),
         checkpoint_every = 1,
         checkpoint_folder = './checkpoints',
+        eval_results_folder = './eval-results',
         decay_lr_factor = 0.5,
         decay_lr_if_not_improved_steps = 3,    # decay learning rate if validation loss does not improve for this amount of epochs
         early_stop_if_not_improved_steps = 10, # they do early stopping if 10 evals without improved loss
@@ -298,6 +301,7 @@ class Trainer(Module):
         self.accelerator = Accelerator(
             cpu = cpu,
             gradient_accumulation_steps = grad_accum_every,
+            log_with = 'wandb' if use_wandb else None,
             **accelerate_kwargs
         )
 
@@ -306,6 +310,7 @@ class Trainer(Module):
         # have the trainer detect if the model is stereo and handle the data accordingly
 
         self.model_is_stereo = model.stereo
+        self.sample_rate = model.sample_rate
 
         # epochs
 
@@ -334,16 +339,37 @@ class Trainer(Module):
             # give musdb special treatment
             dataset = MusDBDataset(dataset)
 
+        # maybe split dataset for eval
+
+        need_split_train_dataset = random_split_dataset_for_eval_frac > 0
+        assert not (need_split_train_dataset and exists(eval_dataset)), f'`eval_dataset` must not be passed in if `random_split_dataset_for_eval_frac` set greater than 0.'
+
+        if not exists(eval_dataset) and need_split_train_dataset:
+            train_size = int((1. - random_split_dataset_for_eval_frac) * len(dataset))
+            eval_size = len(dataset) - train_size
+            dataset, eval_dataset = random_split(dataset, (train_size, eval_size))
+
+        # print
+
+        self.print(f'\ntraining on dataset of {len(dataset)} samples')
+
         # torch from this point on
 
         dataset = CastTorch(dataset, device = device)
 
+        if exists(eval_dataset):
+            eval_dataset = CastTorch(eval_dataset, device = device)
+
         # handle model is not stereo but data is stereo
 
         if not self.model_is_stereo:
             dataset = StereoToMonoDataset(dataset)
 
+            if exists(eval_dataset):
+                eval_dataset = StereoToMonoDataset(eval_dataset)
+
         # augmentations
+        # only for training `dataset`
 
         if augment_gain:
             dataset = GainAugmentation(dataset)
@@ -358,8 +384,18 @@ class Trainer(Module):
         dataloader = DataLoader(dataset, batch_size = batch_size, drop_last = True, shuffle = True, collate_fn = collate_fn)
 
         eval_dataloader = None
+
         if exists(eval_dataset):
-            eval_dataloader = DataLoader(eval_dataset, batch_size = batch_size, drop_last = True, shuffle = True, collate_fn = collate_fn)
+            eval_dataloader = DataLoader(eval_dataset, batch_size = batch_size, drop_last = False, shuffle = True, collate_fn = default_collate_fn)
+
+        self.eval_dataset = eval_dataset
+
+        # evaluation results
+
+        eval_results_folder = Path(eval_results_folder)
+        eval_results_folder.mkdir(parents = True, exist_ok = True)
+
+        self.eval_results_folder = eval_results_folder
 
         # maybe experiment tracker
 
@@ -376,6 +412,8 @@ class Trainer(Module):
                 wandb_tracker = self.accelerator.trackers[0]
                 wandb_tracker.run.name = experiment_run_name
 
+        self.use_wandb = use_wandb
+
         # decay lr logic
 
         scheduler = StepLR(optimizer, 1, gamma = decay_lr_factor)
@@ -423,12 +461,23 @@ class Trainer(Module):
         # prepare eval
 
         if self.needs_eval:
+            self.print(f'\nevaluating on dataset of {len(eval_dataset)} samples')
+
             self.eval_dataloader = self.accelerator.prepare(eval_dataloader)
 
         # step
 
         self.register_buffer('step', tensor(0))
 
+        self.register_buffer('zero', tensor(0.), persistent = False)
+
+    def clear_folders(self):
+        rmtree(str(self.checkpoint_folder), ignore_errors = True)
+        self.checkpoint_folder.mkdir(parents = True, exist_ok = True)
+
+        rmtree(str(self.eval_results_folder), ignore_errors = True)
+        self.eval_results_folder.mkdir(parents = True, exist_ok = True)
+
     @property
     def device(self):
         return self.accelerator.device
@@ -452,10 +501,12 @@ class Trainer(Module):
 
     def forward(self):
 
+        self.print(f'\nstarting training for HSTasNet with {self.unwrapped_model.num_parameters} params\n')
+
         exceeds_max_step = False
         past_eval_losses = [] # for learning rate decay and early stopping detection
 
-        for epoch in range(self.max_epochs):
+        for epoch in range(1, self.max_epochs + 1):
 
             self.model.train()
 
@@ -504,30 +555,70 @@ class Trainer(Module):
 
                 # evaluation at the end of each epoch
 
-                eval_losses = []
+                avg_eval_loss = self.zero
 
+                eval_losses = []
                 for eval_audio, eval_targets, eval_audio_lens in self.eval_dataloader:
 
                     with torch.no_grad():
                         self.model.eval()
 
                         eval_loss = self.model(
-                            audio,
-                            targets = targets,
+                            eval_audio,
+                            targets = eval_targets,
                             audio_lens = eval_audio_lens,
                             auto_curtail_length_to_multiple = True
                         )
 
                         eval_losses.append(eval_loss)
 
-                    avg_eval_loss = stack(eval_losses).mean()
-                    past_eval_losses.append(avg_eval_loss)
+                avg_eval_loss = stack(eval_losses).mean()
+                past_eval_losses.append(avg_eval_loss)
 
                 self.print(f'[{epoch}] eval loss: {avg_eval_loss.item():.3f}')
 
-                self.log(
-                    valid_loss = avg_eval_loss
-                )
+                eval_logs = dict(valid_loss = avg_eval_loss)
+
+                if self.is_main:
+                    model = self.unwrapped_model
+
+                    # take a random sample from eval dataset and store the audio and spectrogram results
+
+                    rand_index = randrange(len(self.eval_dataset))
+                    eval_audio, *_ = self.eval_dataset[rand_index]
+
+                    with torch.no_grad():
+                        model.eval()
+
+                        eval_audio = rearrange(eval_audio, '... -> 1 ...')
+                        separated_audio, _ = model(eval_audio)
+                        separated_audio = rearrange(separated_audio, '1 ... -> ...')
+
+                    # make sure folder exists - each evaluation epoch gets a folder for separated audio and spec
+
+                    one_eval_result_folder = self.eval_results_folder / str(epoch)
+                    one_eval_result_folder.mkdir(parents = True, exist_ok = True)
+
+                    # save files to folder
+
+                    eval_audio_paths = []
+
+                    model.save_tensor_to_file(one_eval_result_folder / 'audio.mp3', eval_audio, overwrite = True)
+
+                    eval_audio_paths.append(('eval_audio', str(one_eval_result_folder / 'audio.mp3')))
+
+                    for index, target_audio in enumerate(separated_audio):
+                        model.save_tensor_to_file(one_eval_result_folder / f'target.separated.{index}.mp3', target_audio, overwrite = True)
+
+                        eval_audio_paths.append((f'target_audio_{index}', str(one_eval_result_folder / f'{index}.mp3')))
+
+                    if self.use_wandb:
+                        from wandb import Audio
+
+                        for name, eval_audio_path in eval_audio_paths:
+                            eval_logs[name] = Audio(eval_audio_path, sample_rate = self.sample_rate)
+
+                self.log(**eval_logs)
 
             # maybe save
 
